﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows.Interactivity;
using System.Windows;
using System.Windows.Input;
using System.Windows.Threading;

namespace TouchGestureLibrary
{
    public enum TapGestureMode
    {
        Regular,
        Long
    }

    public class TapGestureTrigger : TriggerBase<UIElement>
    {
        #region Class members

        Dictionary<TouchDevice, TapGestureEngine> tapStatuses;

        #endregion

        #region Properties

        private TapGestureMode _mode;
        public TapGestureMode Mode
        {
            get
            {
                return _mode;
            }
            set
            {
                _mode = value;
                if (_mode == TapGestureMode.Long)
                {
                    MinMilliseconds = 200;
                    MaxMilliseconds = 5000;
                }
                else
                {
                    MinMilliseconds = 0;
                    MaxMilliseconds = 1000;
                }
            }
        }
        
        private int _minMilliseconds = 0;
        public int MinMilliseconds
        {
            get
            {
                return _minMilliseconds;
            }
            set
            {
                _minMilliseconds = value;
            }
        }
        
        private int _maxMilliseconds = 1000;
        public int MaxMilliseconds
        {
            get
            {
                return _maxMilliseconds;
            }
            set
            {
                _maxMilliseconds = value;
            }
        }

        private double _maxMovement = 25;
        public double MaxMovement
        {
            get
            {
                return _maxMovement;
            }
            set
            {
                _maxMovement = value;
            }
        }

        #endregion
        
        #region Events

        public event EventHandler Tap;

        private void OnTap()
        {
            this.InvokeActions(null);
            if (Tap != null)
            {
                Tap(this.AssociatedObject, EventArgs.Empty);
            }
        }

        #endregion

        #region Constructors

        public TapGestureTrigger()
            : base()
        {
            tapStatuses = new Dictionary<TouchDevice, TapGestureEngine>();
        }

        #endregion

        #region Overridden methods

        protected override void OnAttached()
        {
            base.OnAttached();
            this.AssociatedObject.TouchDown += new EventHandler<TouchEventArgs>(AssociatedObject_TouchDown);
            this.AssociatedObject.TouchMove += new EventHandler<TouchEventArgs>(AssociatedObject_TouchMove);
            this.AssociatedObject.TouchUp += new EventHandler<TouchEventArgs>(AssociatedObject_TouchUp);
        }
        
        #endregion

        #region Event handlers
        
        void AssociatedObject_TouchDown(object sender, TouchEventArgs e)
        {
            //e.Handled = true;
            RemoveGesture(e.TouchDevice);

            TapGestureEngine gesture = new TapGestureEngine(MinMilliseconds, MaxMilliseconds, MaxMovement);
            gesture.TouchDevice = e.TouchDevice;
            gesture.GestureAborted += new EventHandler(gesture_GestureAborted);
            gesture.GestureStarted += new EventHandler(gesture_GestureStarted);
            gesture.GestureCompleted += new EventHandler(gesture_GestureCompleted);

            DateTime dt = TimestampToDateTime(e.Timestamp);

            Point position = e.GetTouchPoint(AssociatedObject).Position;

            gesture.TrackTouchDown(dt, position);
            tapStatuses.Add(e.TouchDevice, gesture);

        }

        void AssociatedObject_TouchMove(object sender, TouchEventArgs e)
        {
            //e.Handled = true;
            if (!tapStatuses.ContainsKey(e.TouchDevice))
            {
                return;
            }
            DateTime dt = TimestampToDateTime(e.Timestamp);

            Point position = e.GetTouchPoint(AssociatedObject).Position;

            tapStatuses[e.TouchDevice].TrackTouchMove(dt, position);
        }

        void AssociatedObject_TouchUp(object sender, TouchEventArgs e)
        {
            //e.Handled = true;
            if (!tapStatuses.ContainsKey(e.TouchDevice))
            {
                return;
            }
            DateTime dt = TimestampToDateTime(e.Timestamp);

            Point position = e.GetTouchPoint(AssociatedObject).Position;

            tapStatuses[e.TouchDevice].TrackTouchUp(dt, position);
        }

        private static DateTime TimestampToDateTime(int timestamp)
        {

            DateTime dt = DateTime.Now;
            dt.AddMilliseconds(Environment.TickCount - timestamp);
            return dt;
        }

        private void RemoveGesture(TouchDevice device)
        {
            if (!tapStatuses.ContainsKey(device))
            {
                return;
            }
            tapStatuses[device].AbortGesture();
            tapStatuses.Remove(device);
        }
        
        #endregion

        #region Gesture events

        void gesture_GestureCompleted(object sender, EventArgs e)
        {
            TapGestureEngine gesture = sender as TapGestureEngine;
            if (gesture == null)
                return;

            RemoveGesture(gesture.TouchDevice);

            OnTap();
        }

        void gesture_GestureStarted(object sender, EventArgs e)
        {     
        }

        void gesture_GestureAborted(object sender, EventArgs e)
        {
            TapGestureEngine gesture = sender as TapGestureEngine;
            if (gesture == null)
                return;

            if (!tapStatuses.ContainsKey(gesture.TouchDevice))
            {
                return;
            }
            tapStatuses.Remove(gesture.TouchDevice);
        }

        #endregion
    }
}
