﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;
using Blake.NUI.WPF.Utility;
using Microsoft.Surface.Presentation.Controls;
using Blake.NUI.WPF.ManipulationHelpers;

namespace BoundarySlider
{
    public partial class Boundary2DSlider : UserControl
    {
        BoundedManipulationRange horizontalRange;
        BoundedManipulationRange verticalRange;

        public Boundary2DSlider()
        {
            InitializeComponent();

            horizontalRange = new BoundedManipulationRange()
                {
                    ElasticMargin = 40,
                    Position = thumbTranslate.X,
                    LowerBoundary = 0
                };

            verticalRange = new BoundedManipulationRange()
                {
                    ElasticMargin = 40,
                    Position = thumbTranslate.Y,
                    LowerBoundary = 0
                };
        }

        private void thumb_ManipulationStarting(object sender, ManipulationStartingEventArgs e)
        {
            e.ManipulationContainer = container;
            e.Handled = true;
        }

        private void thumb_ManipulationDelta(object sender, ManipulationDeltaEventArgs e)
        {
            UpdateRanges();

            horizontalRange.Position += e.DeltaManipulation.Translation.X;
            verticalRange.Position += e.DeltaManipulation.Translation.Y;

            thumbTranslate.X = horizontalRange.BoundedPosition;
            thumbTranslate.Y = verticalRange.BoundedPosition;

            StartInertiaIfOutOfBoundsDuringInertia(e);

            e.ReportBoundaryFeedback(new ManipulationDelta(new Vector(horizontalRange.ElasticOffset,
                                                                      verticalRange.ElasticOffset),
                                                           0,
                                                           new Vector(),
                                                           new Vector()));

            e.Handled = true;
        }

        private void StartInertiaIfOutOfBoundsDuringInertia(ManipulationDeltaEventArgs e)
        {
            if (e.IsInertial)
            {
                //During inertia trying to push past elasticMargin
                if (horizontalRange.IsPositionBelowLowerElasticMargin &&
                    e.Velocities.LinearVelocity.X < 0)
                {
                    //Restart inertia so we can bounce back immediately instead of waiting
                    e.StartInertia();
                }
                else if (horizontalRange.IsPositionAboveUpperElasticMargin &&
                         e.Velocities.LinearVelocity.X > 0)
                {
                    e.StartInertia();
                }
                else if (verticalRange.IsPositionBelowLowerElasticMargin &&
                         e.Velocities.LinearVelocity.Y < 0)
                {
                    e.StartInertia();
                }
                else if (verticalRange.IsPositionAboveUpperElasticMargin &&
                         e.Velocities.LinearVelocity.Y > 0)
                {
                    e.StartInertia();
                }
            }
        }

        private void thumb_ManipulationBoundaryFeedback(object sender, ManipulationBoundaryFeedbackEventArgs e)
        {
            Vector delta = e.BoundaryFeedback.Translation;

            containerTranslate.X = delta.X;
            containerTranslate.Y = delta.Y;

            e.Handled = true;
        }

        private void thumb_ManipulationInertiaStarting(object sender, ManipulationInertiaStartingEventArgs e)
        {
            if (horizontalRange.IsPositionOutOfBounds ||
                verticalRange.IsPositionOutOfBounds)
            {
                //if we're out of bounds
                Vector velocity = new Vector();
                velocity.X = -2 * Math.Sign(horizontalRange.BoundaryOverflow);
                velocity.Y = -2 * Math.Sign(verticalRange.BoundaryOverflow);

                Vector normalizedVelocity = velocity;
                normalizedVelocity.Normalize();
                
                double displacement = (Math.Abs(horizontalRange.BoundaryOverflow) + 1) * Math.Abs(normalizedVelocity.X) +
                                      (Math.Abs(verticalRange.BoundaryOverflow) + 1) * Math.Abs(normalizedVelocity.Y);
                e.TranslationBehavior.DesiredDisplacement = displacement;

                velocity.X = -Math.Min(2, displacement / 10) * Math.Sign(horizontalRange.BoundaryOverflow);
                velocity.Y = -Math.Min(2, displacement / 10) * Math.Sign(verticalRange.BoundaryOverflow);
                e.TranslationBehavior.InitialVelocity = velocity;
            }
            else
            {
                //otherwise regular inertia within the range
                e.TranslationBehavior.DesiredDeceleration = 0.002;
            }
        }

        private void thumb_ManipulationCompleted(object sender, ManipulationCompletedEventArgs e)
        {
            UpdateRanges();

            if (horizontalRange.IsPositionOutOfBounds ||
                verticalRange.IsPositionOutOfBounds)
            {
                //Manipulation has completed but position is beyond unstretch range.
                //This can happen when inertia completes after flicking the object past the boundary.
                //Use this helper to restart the manipulation so inertia logic can take over.
                InertiaDummyManipulator.ManipulateAndStartInertia(thumb);
            }
        }

        private void UpdateRanges()
        {
            horizontalRange.UpperBoundary = container.ActualWidth - thumb.ActualWidth;
            verticalRange.UpperBoundary = container.ActualHeight - thumb.ActualHeight;
        }
    }
}
