﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;
using Blake.NUI.WPF.ManipulationHelpers;
using Blake.NUI.WPF.Utility;
using Microsoft.Surface.Presentation.Input;

namespace NestedManipulations.Controls
{
    public class RectangleManipulationContentControl : ContentControl, INotifyParentManipulationDelta
    {
        #region Properties

        #region IgnoreTags

        public bool IgnoreTags
        {
            get { return (bool)GetValue(IgnoreTagsProperty); }
            set { SetValue(IgnoreTagsProperty, value); }
        }

        // Using a DependencyProperty as the backing store for IgnoreTags.  This enables animation, styling, binding, etc...
        public static readonly DependencyProperty IgnoreTagsProperty =
            DependencyProperty.Register("IgnoreTags", typeof(bool), typeof(RectangleManipulationContentControl), new UIPropertyMetadata(false));

        #endregion

        #region ChildFeedback

        protected Vector ChildFeedback = new Vector();

        #endregion

        #region Translate

        private TranslateTransform _translate;
        protected TranslateTransform Translate
        {
            get
            {
                if (_translate == null)
                    _translate = new TranslateTransform();
                return _translate;
            }
        }

        #endregion

        #region ManipulationRanges

        BoundedManipulationRange HorizontalRange { get; set; }
        BoundedManipulationRange VerticalRange { get; set; }

        #endregion

        #endregion

        #region Template Parts

        private const string Border_name = "PART_Border";

        private Border _border;
        protected Border Border
        {
            get
            {
                return _border;
            }
            set
            {
                if (_border == value)
                    return;

                if (_border != null)
                {
                    _border.ManipulationStarting -= border_ManipulationStarting;
                    _border.ManipulationDelta -= border_ManipulationDelta;
                    _border.ManipulationInertiaStarting -= border_ManipulationInertiaStarting;
                    _border.ManipulationBoundaryFeedback -= border_ManipulationBoundaryFeedback;
                    _border.ManipulationCompleted += border_ManipulationCompleted;
                    _border.GotTouchCapture += border_GotTouchCapture;
                    _border.RenderTransform = null;
                }

                _border = value;

                if (_border != null)
                {
                    _border.IsManipulationEnabled = true;
                    _border.ManipulationStarting += border_ManipulationStarting;
                    _border.ManipulationDelta += border_ManipulationDelta;
                    _border.ManipulationInertiaStarting += border_ManipulationInertiaStarting;
                    _border.ManipulationBoundaryFeedback += border_ManipulationBoundaryFeedback;
                    _border.ManipulationCompleted += border_ManipulationCompleted;
                    _border.GotTouchCapture += border_GotTouchCapture;
                    _border.RenderTransform = Translate;
                }
            }
        }

        #endregion

        #region Constructors

        static RectangleManipulationContentControl()
        {
            DefaultStyleKeyProperty.OverrideMetadata(typeof(RectangleManipulationContentControl), new FrameworkPropertyMetadata(typeof(RectangleManipulationContentControl)));
        }

        public RectangleManipulationContentControl()
        {
            this.HorizontalAlignment = System.Windows.HorizontalAlignment.Left;
            this.VerticalAlignment = System.Windows.VerticalAlignment.Top;

            HorizontalRange = new BoundedManipulationRange()
                {
                    ElasticMargin = 40,
                    Position = Translate.X
                };

            VerticalRange = new BoundedManipulationRange()
                {
                    ElasticMargin = 40,
                    Position = Translate.Y
                };
        }

        #endregion

        #region Overridden Methods

        public override void OnApplyTemplate()
        {
            base.OnApplyTemplate();

            Border = this.GetTemplateChild(Border_name) as Border;
        }

        #endregion

        #region Public Methods

        #region INotifyParentManipulationDelta

        public void ParentManipulationDelta(object sender, ManipulationDeltaEventArgs e)
        {
            if (Manipulation.IsManipulationActive(Border))
            {
                var delta = e.DeltaManipulation;

                UpdateRanges();

                HorizontalRange.Position += -delta.Translation.X;
                VerticalRange.Position += -delta.Translation.Y;

                Translate.X = HorizontalRange.BoundedPosition;
                Translate.Y = VerticalRange.BoundedPosition;
            }

            NotifyChildManipulationDelta(e);
        }

        #endregion

        #endregion

        #region Private Methods

        void border_ManipulationStarting(object sender, ManipulationStartingEventArgs e)
        {
            if (e.OriginalSource != Border)
                return;
            e.Handled = true;

            e.ManipulationContainer = this.Parent as IInputElement;
        }

        void border_ManipulationDelta(object sender, ManipulationDeltaEventArgs e)
        {
            if (e.OriginalSource != Border)
                return;
            e.Handled = true;

            HorizontalRange.Position += e.DeltaManipulation.Translation.X;
            VerticalRange.Position += e.DeltaManipulation.Translation.Y;

            UpdateTranslation();

            StartInertiaIfOutOfBoundsDuringInertia(e);

            e.ReportBoundaryFeedback(new ManipulationDelta(new Vector(HorizontalRange.ElasticOffset,
                                                                      VerticalRange.ElasticOffset),
                                                           0,
                                                           new Vector(),
                                                           new Vector()));

            NotifyChildManipulationDelta(e);
        }

        void border_ManipulationInertiaStarting(object sender, ManipulationInertiaStartingEventArgs e)
        {
            if (e.OriginalSource != Border)
                return;
            e.Handled = true;

            if (HorizontalRange.IsPositionOutOfBounds ||
                VerticalRange.IsPositionOutOfBounds)
            {
                double vel = 1;
                //if we're out of bounds
                Vector velocity = new Vector();
                velocity.X = -vel * Math.Sign(HorizontalRange.BoundaryOverflow);
                velocity.Y = -vel * Math.Sign(VerticalRange.BoundaryOverflow);

                Vector normalizedVelocity = velocity;
                normalizedVelocity.Normalize();

                double displacement = (Math.Abs(HorizontalRange.BoundaryOverflow) + 1) * Math.Abs(normalizedVelocity.X) +
                                      (Math.Abs(VerticalRange.BoundaryOverflow) + 1) * Math.Abs(normalizedVelocity.Y);
                e.TranslationBehavior.DesiredDisplacement = displacement;

                velocity.X = -Math.Min(vel, displacement / 10) * Math.Sign(HorizontalRange.BoundaryOverflow);
                velocity.Y = -Math.Min(vel, displacement / 10) * Math.Sign(VerticalRange.BoundaryOverflow);
                e.TranslationBehavior.InitialVelocity = velocity;
            }
            else
            {
                //otherwise regular inertia within the range
                e.TranslationBehavior.DesiredDeceleration = 0.002;
            }
        }

        void border_ManipulationCompleted(object sender, ManipulationCompletedEventArgs e)
        {
            if (e.OriginalSource != Border)
                return;
            e.Handled = true;

            UpdateRanges();

            if (HorizontalRange.IsPositionOutOfBounds ||
                VerticalRange.IsPositionOutOfBounds)
            {
                //Manipulation has completed but position is beyond unstretch range.
                //This can happen when inertia completes after flicking the object past the boundary.
                //Use this helper to restart the manipulation so inertia logic can take over.
                InertiaDummyManipulator.ManipulateAndStartInertia(Border);
            }
        }

        void border_ManipulationBoundaryFeedback(object sender, ManipulationBoundaryFeedbackEventArgs e)
        {
            //Do not handle our own boundary feedback, let it bubble up
            if (e.OriginalSource == this.Border)
                return;
            e.Handled = true;
            ChildFeedback.X = e.BoundaryFeedback.Translation.X;
            ChildFeedback.Y = e.BoundaryFeedback.Translation.Y;

            UpdateTranslation();
        }

        void border_GotTouchCapture(object sender, TouchEventArgs e)
        {
            if (IsTag(e.TouchDevice) && IgnoreTags)
            {
                e.TouchDevice.Capture(Border);
                ManipulationUtility.CaptureTouchDeviceToManipulationEnabledParent(this, e.TouchDevice);
            }
        }

        bool IsTag(TouchDevice device)
        {
            //Surface SDK 2.0 stuff goes here
            //NDA for now...sorry!
            return false;
        }

        private void UpdateTranslation()
        {
            UpdateRanges();
            Translate.X = HorizontalRange.BoundedPosition + ChildFeedback.X;
            Translate.Y = VerticalRange.BoundedPosition + ChildFeedback.Y;
        }

        private void UpdateRanges()
        {
            Size parentSize = new Size(100, 100);
            Thickness parentBorder = new Thickness(0);

            FrameworkElement parentElement = this.Parent as FrameworkElement;
            if (parentElement != null)
            {
                parentSize = parentElement.RenderSize;
            }

            Control parentControl = this.Parent as Control;
            if (parentControl != null)
            {
                parentBorder = parentControl.BorderThickness;
            }

            HorizontalRange.LowerBoundary = 0;
            HorizontalRange.UpperBoundary = parentSize.Width - this.ActualWidth - parentBorder.Right - this.BorderThickness.Right;

            VerticalRange.LowerBoundary = 0;
            VerticalRange.UpperBoundary = parentSize.Height - this.ActualHeight - parentBorder.Bottom - this.BorderThickness.Bottom;
        }

        private void StartInertiaIfOutOfBoundsDuringInertia(ManipulationDeltaEventArgs e)
        {
            if (e.IsInertial)
            {
                //During inertia trying to push past elasticMargin
                if (HorizontalRange.IsPositionBelowLowerElasticMargin &&
                    e.Velocities.LinearVelocity.X < 0)
                {
                    //Restart inertia so we can bounce back immediately instead of waiting
                    e.StartInertia();
                }
                else if (HorizontalRange.IsPositionAboveUpperElasticMargin &&
                         e.Velocities.LinearVelocity.X > 0)
                {
                    e.StartInertia();
                }
                else if (VerticalRange.IsPositionBelowLowerElasticMargin &&
                         e.Velocities.LinearVelocity.Y < 0)
                {
                    e.StartInertia();
                }
                else if (VerticalRange.IsPositionAboveUpperElasticMargin &&
                         e.Velocities.LinearVelocity.Y > 0)
                {
                    e.StartInertia();
                }
            }
        }

        void NotifyChildManipulationDelta(ManipulationDeltaEventArgs e)
        {
            var obj = this.Content as INotifyParentManipulationDelta;
            if (obj == null)
                return;

            obj.ParentManipulationDelta(this, e);
        }

        #endregion
    }
}
